/*
  Copyright 2021 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.partner.friend.view

import android.app.Activity
import android.content.Context.INPUT_METHOD_SERVICE
import android.content.res.Configuration
import android.os.Bundle
import android.text.Editable
import android.text.TextWatcher
import android.util.TypedValue
import android.view.Gravity
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.view.inputmethod.InputMethodManager
import android.widget.LinearLayout
import android.widget.Toast
import androidx.constraintlayout.widget.ConstraintSet
import androidx.constraintlayout.widget.ConstraintSet.*
import androidx.core.graphics.drawable.DrawableCompat
import androidx.fragment.app.Fragment
import androidx.lifecycle.Observer
import androidx.lifecycle.ViewModelProvider
import androidx.recyclerview.widget.RecyclerView
import androidx.recyclerview.widget.SimpleItemAnimator
import com.kakao.sdk.common.model.ClientError
import com.kakao.sdk.common.model.ClientErrorCause
import com.kakao.sdk.common.model.KakaoSdkError
import com.kakao.sdk.partner.friend.Constants
import com.kakao.sdk.partner.friend.adapter.FriendsListAdapter
import com.kakao.sdk.partner.friend.adapter.PickedFriendsAdapter
import com.kakao.sdk.partner.friend.model.*
import com.kakao.sdk.partner.friend.util.LayoutUtils
import com.kakao.sdk.partner.friend.viewmodel.FriendViewModel
import com.kakao.sdk.partner.friend.viewmodel.SearchStatus
import com.kakao.sdk.v2.partner.friend.R
import com.kakao.sdk.v2.partner.friend.databinding.*
import kotlin.math.max

/**
 * @suppress
 */
internal class FriendPickerFragment : Fragment() {
    private var _binding: KakaoSdkFragmentFriendPickerBinding? = null
    private val binding get() = _binding!!

    private var title = ""
    private var friendOrder: PickerFriendOrder? = null
    private var isSingle = false
    private var isPopup = false
    private var enableSearch = true
    private var enableIndex = true
    private var searchPlaceHolderText = ""
    private var showPickedFriend = true
    private var showMyProfile = true
    private var showFavorite = true
    private var maxPickableCount = 1
    private var minPickableCount = 1

    private var titleBar: KakaoSdkTitleBarBinding? = null
    private var pickedFriendsRv: KakaoSdkPickedFriendsViewBinding? = null
    private var searchBarLayout: KakaoSdkSearchBarBinding? = null
    private var friendsRv: KakaoSdkFriendsListBinding? = null

    private val pickedFriendsAdapter: PickedFriendsAdapter by lazy {
        PickedFriendsAdapter(viewModel.pickedFriends.value?.toMutableList() ?: mutableListOf()) {
            friendListAdapter.uncheckFriend(it)
            viewModel.removeUncheckedFriend(it)
        }
    }

    private val friendListAdapter: FriendsListAdapter by lazy {
        FriendsListAdapter(
            isPopup, isSingle, enableSearch, showPickedFriend, maxPickableCount,
            checkCallback = {
                if (it == null) {
                    // 선택 가능한 숫자 이상으로 친구를 선택한 경우 null 리턴 받음
                    Toast.makeText(
                        context,
                        getString(R.string.max_pickable_count_message, maxPickableCount),
                        Toast.LENGTH_SHORT
                    ).show()
                } else {
                    // 정상적으로 친구가 선택된 경우
                    pickedFriendsAdapter.addPickedFriend(it)
                    pickedFriendsRv?.root?.scrollToPosition(0) // 최근 선택된 친구가 보이도록 스크롤 이동
                    friendListAdapter.checkFriend(it) // 친구 리스트 체크박스 true로 변경
                    viewModel.addCheckedFriend(it)
                }
            },
            uncheckCallback = {
                pickedFriendsAdapter.removePickedFriend(it)
                friendListAdapter.uncheckFriend(it) // 친구 리스트 체크박스 false로 변경
                viewModel.removeUncheckedFriend(it)
            }
        )
    }
    private val viewModel by lazy {
        ViewModelProvider(requireActivity(), ViewModelProvider.NewInstanceFactory()).get(
            FriendViewModel::class.java
        )
    }

    override fun onCreateView(
        inflater: LayoutInflater, container: ViewGroup?,
        savedInstanceState: Bundle?
    ): View {
        _binding = KakaoSdkFragmentFriendPickerBinding.inflate(layoutInflater, container, false)
        return binding.root
    }

    override fun onViewCreated(view: View, savedInstanceState: Bundle?) {
        super.onViewCreated(view, savedInstanceState)
        setProperty()
        createLayout()

        // 최초 뷰 생성시 savedInstanceState가 null이므로 화면 회전 구분
        if (savedInstanceState == null) {
            viewModel.fetchData(requireContext()) // 최초 뷰 생성시에만 데이터 로딩
        }
        initializeView()
        observeData()
    }

    private fun createLayout() {
        binding.mainLayout.apply {
            titleBar = KakaoSdkTitleBarBinding.inflate(layoutInflater, this, false).apply {
                title.text = this@FriendPickerFragment.title

                if (isSingle) {
                    pickedCountTv.visibility = View.INVISIBLE
                }
                addView(root)
            }
            if (!isSingle && showPickedFriend) {
                pickedFriendsRv =
                    KakaoSdkPickedFriendsViewBinding.inflate(layoutInflater, this, false).apply {
                        addView(root)
                    }
            }
            if (enableSearch) {
                searchBarLayout =
                    KakaoSdkSearchBarBinding.inflate(layoutInflater, this, false).apply {
                        searchBar.hint = searchPlaceHolderText
                        addView(root)
                    }
                searchBarLayout?.apply {
                    searchBar.addTextChangedListener(object : TextWatcher {
                        override fun beforeTextChanged(
                            s: CharSequence?,
                            start: Int,
                            count: Int,
                            after: Int
                        ) {
                        }

                        override fun onTextChanged(
                            s: CharSequence?,
                            start: Int,
                            before: Int,
                            count: Int
                        ) {
                        }

                        override fun afterTextChanged(s: Editable?) {
                            if (s.isNullOrEmpty()) {
                                searchBarDeleteBtn.visibility = View.GONE
                                placeholderMarginView.visibility = View.VISIBLE
                            } else {
                                searchBarDeleteBtn.visibility = View.VISIBLE
                                placeholderMarginView.visibility = View.GONE
                            }
                            viewModel.searchFriends(requireContext(), s.toString())
                        }
                    })
                    searchBarDeleteBtn.setOnClickListener {
                        searchBar.setText("")
                    }
                }
            }
            friendsRv = KakaoSdkFriendsListBinding.inflate(layoutInflater, this, false).apply {
                addView(root, LinearLayout.LayoutParams(ViewGroup.LayoutParams.MATCH_PARENT, 0, 1f))

                if (enableIndex && friendOrder == null) {
                    sideIv.recyclerView = friendsList
                    val wrapped = DrawableCompat.wrap(sideIv.background)
                    val typedValue = TypedValue()
                    context.theme.resolveAttribute(R.attr.gray050s, typedValue, true)
                    DrawableCompat.setTint(wrapped.mutate(), typedValue.data)
                    friendsList.addOnScrollListener(sideIv.onScrollListener)
                }
            }
        }

        if (LayoutUtils.orientation != Configuration.ORIENTATION_PORTRAIT) {
            binding.emptyFriendView.emptyFriendImage.visibility = View.GONE
        }

        if (isPopup) {
            setPopupLayout()
        }
    }

    private fun initializeView() {
        pickedFriendsRv?.root?.adapter = pickedFriendsAdapter
        friendsRv?.friendsList?.apply {
            adapter = friendListAdapter
            // 즐겨찾기, 친구 리스트에 중복되는 친구를 터치할 경우 깜빡거리는 현상 제거
            itemAnimator.apply {
                if (this is SimpleItemAnimator) {
                    supportsChangeAnimations = false
                }
            }
            // 스크롤하면 키보드 숨기고 검색바에서 포커스 잃도록 처리
            addOnScrollListener(object : RecyclerView.OnScrollListener() {
                override fun onScrollStateChanged(recyclerView: RecyclerView, newState: Int) {
                    super.onScrollStateChanged(recyclerView, newState)
                    searchBarLayout?.searchBar?.clearFocus()
                    hideKeyboard()
                }
            })
        }
        titleBar?.apply {
            backBtn.setOnClickListener {
                sendError(ClientError(ClientErrorCause.Cancelled, "cancelled."))
            }
            doneBtn.setOnClickListener {
                sendPickedFriends()
            }
        }
    }

    private fun setPopupLayout() {
        titleBar?.apply {
            val marginTop = if (LayoutUtils.orientation == Configuration.ORIENTATION_PORTRAIT) {
                resources.getDimensionPixelSize(R.dimen.popup_title_bar_padding_top)
            } else {
                resources.getDimensionPixelSize(R.dimen.popup_title_bar_padding_top_landscape)
            }
            root.setPadding(0, marginTop, 0, 0)
            title.apply {
                gravity = Gravity.CENTER

                // 뷰가 그려지기 전에 width를 계산
                backBtn.measure(View.MeasureSpec.UNSPECIFIED, View.MeasureSpec.UNSPECIFIED)
                doneBtnLayout.measure(View.MeasureSpec.UNSPECIFIED, View.MeasureSpec.UNSPECIFIED)

                // 타이틀 좌, 우 뷰 중에서 더 넓은 뷰 넓이에 마진값 적용
                val marginHorizontal = max(
                    backBtn.measuredWidth,
                    doneBtnLayout.measuredWidth
                ) + resources.getDimensionPixelSize(R.dimen.popup_title_bar_text_margin_start)

                ConstraintSet().apply {
                    clone(root)
                    connect(id, START, PARENT_ID, START, marginHorizontal)
                    connect(id, END, PARENT_ID, END, marginHorizontal)
                }.applyTo(root)
            }
        }
        friendsRv?.friendsList?.apply {
            val paddingBottom = if (LayoutUtils.orientation == Configuration.ORIENTATION_PORTRAIT) {
                resources.getDimensionPixelSize(R.dimen.popup_friends_list_padding_bottom)
            } else {
                resources.getDimensionPixelSize(R.dimen.popup_friends_list_padding_bottom_landscape)
            }
            setPadding(0, 0, 0, paddingBottom)
        }
    }

    private fun observeData() {
        viewModel.run {
            pickerList.observe(viewLifecycleOwner, Observer {
                friendListAdapter.setFriendsList(it)
                if (enableIndex) {
                    setIndexView(it)
                }
            })
            pickedFriends.observe(viewLifecycleOwner, Observer {
                friendListAdapter.setPickedFriends(it)
                pickedFriendsRv?.root?.visibility = if (it.size > 0) View.VISIBLE else View.GONE
                titleBar?.pickedCountTv?.apply {
                    text = it.size.toString()
                    visibility = if (!isSingle && it.size > 0) View.VISIBLE else View.INVISIBLE
                }
                titleBar?.doneBtn?.isEnabled = it.size > 0
            })
            searchStatus.observe(viewLifecycleOwner, Observer {
                if (it == SearchStatus.SEARCHING) {
                    friendsRv?.apply {
                        // 검색할 때 스크롤 최상단으로 이동
                        friendsList.scrollToPosition(0)
                        friendsList.removeOnScrollListener(sideIv.onScrollListener)
                    }
                } else if (it == SearchStatus.DONE) {
                    friendsRv?.apply {
                        friendsList.scrollToPosition(0)
                        friendsList.addOnScrollListener(sideIv.onScrollListener)
                    }
                }
            })
            isError.observe(viewLifecycleOwner, Observer {
                if (it) {
                    titleBar?.doneBtn?.visibility = View.GONE
                    searchBarLayout?.root?.visibility = View.GONE
                    binding.emptyFriendView.root.visibility = View.VISIBLE
                } else {
                    titleBar?.doneBtn?.visibility = View.VISIBLE
                    searchBarLayout?.root?.visibility = View.VISIBLE
                    binding.emptyFriendView.root.visibility = View.GONE
                }
            })
        }
    }

    private fun setProperty() {
        isSingle = viewModel.isSingle
        isPopup = viewModel.isPopup
        viewModel.param.also {
            title = it.title ?: getString(R.string.invite_friends)
            friendOrder = it.pickerFriendOrder
            enableSearch = it.enableSearch ?: Constants.DEFAULT_ENABLE_SEARCH
            enableIndex = it.enableIndex ?: Constants.DEFAULT_ENABLE_INDEX
            searchPlaceHolderText = it.searchPlaceHolderText ?: getString(R.string.search_bar_hint)
            showMyProfile = it.showMyProfile ?: Constants.DEFAULT_SHOW_MY_PROFILE
            showFavorite = it.showFavorite ?: Constants.DEFAULT_SHOW_FAVORITE
            minPickableCount =
                if (isSingle) 1 else it.minPickableCount ?: Constants.DEFAULT_MIN_PICKABLE_COUNT
            maxPickableCount =
                if (isSingle) 1 else it.maxPickableCount ?: Constants.DEFAULT_MAX_PICKABLE_COUNT
            showPickedFriend = if (isSingle) false else it.showPickedFriend
                ?: Constants.DEFAULT_SHOW_PICKED_FRIENDS
        }
    }

    private fun setIndexView(pickerList: MutableList<Item>) {
        val friends = viewModel.friendsList
        var friendListOffset = 0
        var favoriteOffset = -1

        pickerList.forEach { item ->
            ++friendListOffset
            if (item is Item.Header && item.title == getString(R.string.favorite)) {
                favoriteOffset = friendListOffset - 1
                return@forEach
            }
        }

        // extraOffset = RecyclerView 전체 리스트 - 순수 친구 리스트 - favoriteOffset
        val extraOffset = pickerList.size - friends.size - favoriteOffset

        friendsRv?.sideIv?.apply {
            setDataSource(friends, friends.size, favoriteOffset, extraOffset)
            // 팝업 유무에 따라서 인덱스 리스트 변경
            if (isPopup) {
                setIndexItems(R.array.popup_side_indexer, R.array.popup_side_indexer_landscape)
            } else {
                setIndexItems(R.array.side_indexer, R.array.side_indexer_landscape)
            }
        }
    }

    private fun hideKeyboard() {
        val inputMethodManager =
            context?.getSystemService(INPUT_METHOD_SERVICE) as InputMethodManager
        inputMethodManager.hideSoftInputFromWindow(searchBarLayout?.searchBar?.windowToken, 0)
    }

    override fun onDestroyView() {
        super.onDestroyView()
        _binding = null
    }

    private fun sendPickedFriends() {
        if (viewModel.pickedFriends.value?.size ?: 0 < minPickableCount) {
            Toast.makeText(
                requireContext(),
                getString(R.string.min_pickable_count_message, minPickableCount),
                Toast.LENGTH_LONG
            ).show()
        } else {
            viewModel.sendResult()
            requireActivity().finish()
        }
    }

    private fun sendError(exception: KakaoSdkError) {
        viewModel.resultReceiver.send(
            Activity.RESULT_CANCELED,
            Bundle().apply { putSerializable(Constants.KEY_EXCEPTION, exception) }
        )
        requireActivity().finish()
    }
}
